@@ -1,8 +1,8 @@
 /*
- * Vtable-Dumper 1.1
+ * Vtable-Dumper 1.2
  * A tool to list content of virtual tables in a shared library
  * 
- * Copyright (C) 2013-2015 Andrey Ponomarenko's ABI laboratory
+ * Copyright (C) 2013-2016 Andrey Ponomarenko's ABI laboratory
  * 
  * Written by Andrey Ponomarenko
  * 
@@ -33,7 +33,7 @@
 
 #include "dump-vtable.h"
 
-const char* TOOL_VERSION = "1.1";
+const char* TOOL_VERSION = "1.2";
 
 int opt_mangled = 0;
 int opt_demangled = 0;
@@ -43,9 +43,9 @@ int get_VTables(char* file, vtable_info*** vtables)
     int fd;
     struct stat elf_stats;
     
-    int symbol_count;
-    int num;
-    int i;
+    size_t symbol_count;
+    size_t num;
+    size_t i;
     char *sname, *ssname;
     vtable_info* vt;
     
@@ -73,17 +73,50 @@ int get_VTables(char* file, vtable_info*** vtables)
         printf("WARNING Elf Library is out of date!\n");
     }
     
-    elf = elf_begin(fd, ELF_C_READ, NULL);
-    
+    if ((elf = elf_begin(fd, ELF_C_READ, NULL)) == NULL)
+    {
+        return ERR;
+    }
+
     while ((scn = elf_nextscn(elf, scn)) != NULL)
     {
-        gelf_getshdr(scn, &shdr);
+        if (gelf_getshdr(scn, &shdr) != &shdr)
+        {
+            return ERR;
+        }
         
         if (shdr.sh_type == SHT_DYNSYM)
         {
             edata = elf_getdata(scn, edata);
+            if (edata == NULL)
+            {
+                return ERR;
+            }
             
+            // Prevent divide by zero.
+            if (shdr.sh_entsize == 0)
+            {
+                return ERR;
+            }
+
+            // This could truncate on a 32bit machine inspecting a 64bit binary.
             symbol_count = shdr.sh_size / shdr.sh_entsize;
+
+            // Check for potential multiplication wraparound.
+            if (symbol_count > SIZE_MAX / sizeof(vtable_info*))
+            {
+                return ERR;
+            }
+
+            // Add one to symbol_count for NULL storage later.
+            if (SIZE_MAX - 1 < symbol_count)
+            {
+                return ERR;
+            }
+            else 
+            {
+                symbol_count += 1;
+            }
             *vtables = (vtable_info**)malloc(sizeof(vtable_info*) * symbol_count);
             
             num = 0;
@@ -95,9 +128,19 @@ int get_VTables(char* file, vtable_info*** vtables)
                 { // UND
                     continue;
                 }
-                
+
                 sname = elf_strptr(elf, shdr.sh_link, sym.st_name);
+                if (sname == NULL)
+                {
+                    return ERR;
+                }
+
                 ssname = (char*)malloc(1 + strlen(sname));
+                if (ssname == NULL)
+                {
+                    return ERR;
+                }
+
                 strcpy(ssname, sname);
                 
                 if (strstr(ssname, "_ZTV"))
@@ -117,15 +160,15 @@ int get_VTables(char* file, vtable_info*** vtables)
             (*vtables)[num] = NULL;
         }
     }
+
+    elf_end(elf);
+    close(fd);
     
     if (*vtables == NULL)
     {
         return ERR;
     }
     
-    elf_end(elf);
-    close(fd);
-    
     return 0;
 }
 
@@ -324,6 +367,17 @@ void print_VTable(void *dlhndl, vtable_info *vtable)
     printf("\n\n");
 }
 
+void free_vtables(vtable_info **vtables)
+{
+    size_t i;
+    for (i = 0; vtables[i] != NULL; i++)
+    {
+        free(vtables[i]->name);
+        free(vtables[i]);
+    }
+    free(vtables);
+}
+
 int print_Info(char* file)
 {
     int i;
@@ -345,18 +399,16 @@ int print_Info(char* file)
         {
             fprintf(stderr, "%s\n", error);
         }
+        free_vtables(vtables);
         return ERR;
     }
     
     for (i = 0; vtables[i] != NULL; i++)
     {
         print_VTable(dlhndl, vtables[i]);
-        free(vtables[i]->name);
-        free(vtables[i]);
     }
     
-    free(vtables);
-    dlclose(dlhndl);
+    free_vtables(vtables);
     
     return 0;
 }
@@ -365,7 +417,7 @@ void print_Usage()
 {
     printf("Vtable-Dumper %s\n", TOOL_VERSION);
     printf("A tool to list content of virtual tables in a shared library\n");
-    printf("Copyright (C) 2013-2015 Andrey Ponomarenko's ABI laboratory\n");
+    printf("Copyright (C) 2013-2016 Andrey Ponomarenko's ABI laboratory\n");
     printf("License: GNU LGPL or GNU GPL\n\n");
     
     printf("Usage: vtable-dumper [options] object\n");
